using System;
using System.Windows.Forms;
using DicomObjects;
using DicomObjects.EventArguments;
using DicomObjects.DicomUIDs;
using DicomObjects.Enums;

namespace SCU
{
	/// <summary>
	/// Summary description for Form1.
	/// </summary>
	public class Form1 : Form
	{
		private const string CloseOutgoingAssociationLogMessage = "Outgoing Association Closed Normally";
		private const string Node = "localhost";
		private const string CallingAE = "SC_SCU";
		private const string CalledAE = "SC_SCP";
		private const int Port = 104;
		private const string RequestForImage = "Request for Image";
		private const string CommitmentRequest = "added for storage commitment request.";
		private const string SendingStorageCommitmentRequest = "Sending storage commitment request - Transaction UID =";
		private const string IncomingAssocationRequest = "Incoming Assocaiton Request from";
		private const string CommandFieldErrorMessage = "Fatal Error";
		private const string InstanceUIdNotCorrectErrorMessage = "AffectedSOPInstanceUID not Correct";
		private const string ResponseReceivedOnNewAssociationMessage = "Response Received on New Association";
		private const string ResponseReceivedOnOriginalAssociationMessage = "Response Received on Original Association";
		private const string ResponseReceivedOnDifferentAssociationMessage = "Response Received on Different Association";
		private const string StorageCommitmentSucceededMessage = "Storage commitment succeeded for Transaction UID:";
		private const string SucceededImageMessage = "Succeeded Image Instance UID";
		private const string StorageCommitmentFailedMessage = "Storage commitment failed for Transaction UID:";
		private const string FailedImageMessage = "Failed Image Instance UID";
		private const string WrongEventIdReceivedMessage = "Error - Wrong EventID Received";


		internal TextBox TextBox1;
		internal TextBox Report;
		private Button SendSCPushBTN;
		private Timer tmr_dicom_connection;
		private System.ComponentModel.IContainer components;
		private DicomAssociation cn;
		private DicomServer Server;

		public Form1()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.TextBox1 = new TextBox();
			this.Report = new TextBox();
			this.SendSCPushBTN = new Button();
			this.tmr_dicom_connection = new Timer(this.components);
			this.SuspendLayout();
			// 
			// TextBox1
			// 
			this.TextBox1.BackColor = System.Drawing.SystemColors.Control;
			this.TextBox1.BorderStyle = BorderStyle.None;
			this.TextBox1.Font = new System.Drawing.Font("Microsoft Sans Serif", 10F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
			this.TextBox1.Location = new System.Drawing.Point(16, 16);
			this.TextBox1.Multiline = true;
			this.TextBox1.Name = "TextBox1";
			this.TextBox1.ReadOnly = true;
			this.TextBox1.Size = new System.Drawing.Size(712, 72);
			this.TextBox1.TabIndex = 9;
			this.TextBox1.Text = @"This simple example of storage commitment uses the now standard ""push"" model, and as is now regarded as best practice, expects to receive the acknowledgment on a separate association (thoug it will equally happily accept it on the original association if offered).  In order to receive a reversed role association request, it listens on port 105.  The Association is held open for 5 seconds to allow responses on the same connection if the SCP chooses so";
			// 
			// Report
			// 
			this.Report.BorderStyle = BorderStyle.FixedSingle;
			this.Report.Location = new System.Drawing.Point(192, 96);
			this.Report.Multiline = true;
			this.Report.Name = "Report";
			this.Report.ScrollBars = ScrollBars.Vertical;
			this.Report.Size = new System.Drawing.Size(536, 376);
			this.Report.TabIndex = 0;
			this.Report.Text = "";
			// 
			// SendSCPushBTN
			// 
			this.SendSCPushBTN.Location = new System.Drawing.Point(8, 96);
			this.SendSCPushBTN.Name = "SendSCPushBTN";
			this.SendSCPushBTN.Size = new System.Drawing.Size(168, 96);
			this.SendSCPushBTN.TabIndex = 12;
			this.SendSCPushBTN.Text = "Send Storage Commitment Push";
			this.SendSCPushBTN.Click += new System.EventHandler(this.SendSCPushBTN_Click);
			// 
			// tmr_dicom_connection
			// 
			this.tmr_dicom_connection.Interval = 4000;
			this.tmr_dicom_connection.Tick += new System.EventHandler(this.Tmr_dicom_connection_Tick);
			// 
			// Form1
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(744, 486);
			this.Controls.Add(this.SendSCPushBTN);
			this.Controls.Add(this.Report);
			this.Controls.Add(this.TextBox1);
			this.Name = "Form1";
			this.Text = "Storage Commitment VB.NET Sample";
			this.Closing += new System.ComponentModel.CancelEventHandler(this.Form1_Closing);
			this.Load += new System.EventHandler(this.Form1_Load);
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main() 
		{
			Application.Run(new Form1());
		}

		private void Form1_Load(object sender, EventArgs e)
		{
			Server = new DicomServer();
			Server.AssociationRequest +=Server_AssociationRequest;
			Server.N_Event_Report +=Server_N_Event_Report;
			Server.DefaultStatus = 0x110;
			Server.Listen(105);
            // Put the following in back in to enable DicomObjects logging
			// DicomObjects.DicomGlobal.LogToFile("c:\\dicom log files\\", 0x3F);
		}

		private void Form1_Closing(object sender, System.ComponentModel.CancelEventArgs e)
		{
			Server.UnlistenAll();
		}

		private void CloseOutgoingAssociation()
		{
			tmr_dicom_connection.Enabled = false;
			if(cn != null)
			{
				cn.Close();
			}
			cn = null;
			Log(CloseOutgoingAssociationLogMessage);
		}

		private void Tmr_dicom_connection_Tick(object sender, EventArgs e)
		{
			CloseOutgoingAssociation();
		}

        private void Log(string s)
        {
            AppendText(Report, s);
        }

        delegate void SetTextCallback(TextBox box, string text);

        private void AppendText(TextBox box, string text)
        {
            // InvokeRequired required compares the thread ID of the
            // calling thread to the thread ID of the creating thread.
            // If these threads are different, it returns true.
            if (box.InvokeRequired)
            {
                SetTextCallback d = AppendText;
                Invoke(d, new object[] { box, text });
            }
            else
            {
                box.AppendText(text + "\r\n");
            }
        }

		private void SendSCPushBTN_Click(object sender, EventArgs e)
		{

			/* STORAGE COMMITMENT REQUEST
			 
			Action Type/Name		Action Type ID		Attribute		Tag
			
			Request						
			Storage						   1		  TransactionUID (0008,1195)
			Commitment
													  Storage Media
                                                       File-Set ID   (0088,0130)
														
													  Storage Media
													   File-Set UID  (0088,0140)

													   Referenced
													       SOP
													    Sequence     (0008,1199)

													    >Referenced
													     SOP Class
													     UID           >(0008,1150)
														>Referenced
														SOP Instance
														UID			   >(0008,1155)
			 
			 */

			string[] offerTS = new string[2];
            offerTS[0] = TransferSyntaxes.ImplicitVRLittleEndian;
			offerTS[1] = TransferSyntaxes.ExplicitVRLittleEndian;

			// close possible previously opened association and create a new one
			CloseOutgoingAssociation();

			// Enable timer to close outgoing association after a time-out
			tmr_dicom_connection.Enabled = true;
            cn = new DicomAssociation
            {

                // Negotiate a suitable Association with Correct SOP Class
                MetaSOPClass = SOPClasses.StorageCommitmentPush
            };
            cn.RequestedContexts.Add(SOPClasses.StorageCommitmentPush);
			cn.RequestedContexts[1].RequestorSCURole = true;
			cn.RequestedContexts[1].RequestorSCPRole = false;
			cn.RequestedContexts[1].OfferedTS = offerTS;

			cn.Open(Node, Port, CallingAE, CalledAE);

			DicomDataSet ds = new DicomDataSet();
			DicomDataSet referencedImage1 = new DicomDataSet();
			DicomDataSet referencedImage2 = new DicomDataSet();
			DicomDataSetCollection referencedImagesSequence = new DicomDataSetCollection();			

			string instanceUID;

			// the next 3 lines could be repeated N times to send a request for N Images
			referencedImage1.Add(Keyword.ReferencedSOPClassUID, SOPClasses.SecondaryCapture);
			instanceUID = DicomGlobal.NewUID();
			referencedImage1.Add(Keyword.ReferencedSOPInstanceUID, instanceUID);
			referencedImagesSequence.Add(referencedImage1);
			Log($"{RequestForImage} {instanceUID} {CommitmentRequest}");

			// Add another image here
			referencedImage2.Add(Keyword.ReferencedSOPClassUID, SOPClasses.SecondaryCapture);
			instanceUID = DicomGlobal.NewUID();
			referencedImage2.Add(Keyword.ReferencedSOPInstanceUID, instanceUID);
			referencedImagesSequence.Add(referencedImage2);
			Log($"{RequestForImage} {instanceUID} {CommitmentRequest}");

			ds.Add(Keyword.ReferencedSOPSequence, referencedImagesSequence);
			ds.Add(Keyword.TransactionUID, DicomGlobal.NewUID());
			
			Log($"{SendingStorageCommitmentRequest} {ds[Keyword.TransactionUID].Value.ToString()}");

			// Send Storage Commitment Push
			cn.NAction(SOPClasses.StorageCommitmentPush, WellKnownInstances.StorageCommitmentPush,1,ds);
		}

		private void Server_AssociationRequest(object sender, AssociationRequestArgs e)
		{
			Log($"{IncomingAssocationRequest} {e.Association.CallingAET} at {e.Association.RemoteIP}");
		}

		private void Server_N_Event_Report(object sender, N_Event_ReportArgs e)
		{
			string AffectedSOPInstanceUID = Convert.ToString(e.Command[Keyword.AffectedSOPInstanceUID].Value);
			int CommandField = Convert.ToInt32(e.Command[Keyword.CommandField].Value.ToString());
			int status = 0x110;

			if(CommandField != 0x100)
			{
				Log(CommandFieldErrorMessage);
			}
	
			if(AffectedSOPInstanceUID != WellKnownInstances.StorageCommitmentPush)
			{
				status = 0x112; // no such object instance
				Log(InstanceUIdNotCorrectErrorMessage);
			}

			if(cn == null)
			{
				Log(ResponseReceivedOnNewAssociationMessage);
			}
			else
			{
			    Log(cn.AssociationNumber == e.RequestAssociation.AssociationNumber
			            ? ResponseReceivedOnOriginalAssociationMessage
						: ResponseReceivedOnDifferentAssociationMessage);
			}
			
			switch(e.EventID)
			{
				case 1 : // list all successful requests
					Log($"{StorageCommitmentSucceededMessage} {e.Request[Keyword.TransactionUID].Value.ToString()}");
					DicomDataSetCollection dss = (DicomDataSetCollection)e.Request[Keyword.ReferencedSOPSequence].Value;
					foreach(DicomDataSet ds in dss)
					{
                        Log($"{SucceededImageMessage} {ds[Keyword.ReferencedSOPInstanceUID].Value.ToString()}");
					}
					break;
				case 2 : // list all successful requests as well as failed requests
					Log($"{StorageCommitmentFailedMessage} {e.Request[Keyword.TransactionUID].Value.ToString()}");
					DicomDataSetCollection dss1 = (DicomDataSetCollection)e.Request[Keyword.ReferencedSOPSequence].Value;
					foreach(DicomDataSet ds in dss1)
					{
                        Log($"{SucceededImageMessage} {ds[Keyword.ReferencedSOPInstanceUID].Value.ToString()}");
					}
					DicomDataSetCollection dss2 = (DicomDataSetCollection)e.Request[Keyword.FailedSOPSequence].Value;
					foreach(DicomDataSet ds in dss2)
					{
						Log($"{FailedImageMessage} {ds[Keyword.ReferencedSOPInstanceUID].Value.ToString()}");
					}
					break;
				default:
					status = 0x123; // no such action type
					Log(WrongEventIdReceivedMessage);
					break;
			}						
			e.Status = status;
		}
	}
}